/** \file PyTine.h
 *  \brief PyTine Class
 *
 * Implements the bindings between TINE and Python.
 * It is called from the PyTine_py file and talks directly with TINE control system
 *
 */

#ifndef PYTINE_H_
#define PYTINE_H_

#include "tine.h"
#include <iostream>
#include <string>

using namespace std;

/**
 * @brief Implements the bindings between TINE and Python
 */
class PyTine
{
public:

	/**
	 * @brief Returns a pointer to a PyTine object
	 * @return PyTine static pointer
	 */
	static PyTine* instance();

	/**
	 * @brief Client get call
	 * This routine gets the property data from TINE in synchronous mode
	 * @param address string with the TINE address
	 * @param property string with the TINE property
	 * @param size int with the size of the data
	 * @param format int with the format
	 * @param blob void pointer to a data buffer where the data will be stored
	 * @param timeout timeout of the call
	 * @return 0 if successful, otherwise  a TINE completion code which can be interpreted by a call to
	 * @see TINE client ExecLinkEx()
	 * GetLastLinkError().
	 */
	int tget(string address, string property, int size, int format, void * blob, int timeout) const;

	/**
	 * @brief Client put call
	 * This routine sends the given data to TINE
	 * @param address string with the TINE address
	 * @param property string with the TINE property
	 * @param size int with the size of the data
	 * @param format int with the format
	 * @param data void pointer that points to a data buffer with the data to be sent
	 * @return 0 if successful, otherwise  a TINE completion code which can be interpreted by a call to
	 * GetLastLinkError().
	 */
	int tput(string address, string property, int size, int format, void * data) const;

	/**
	 * \brief Client put get call
	 * This routine sends the given data to TINE and receives data from the same property
	 * \param address string with the TINE address
	 * \param property string with the TINE property
	 * \param size int with the size of the data
	 * \param format int with the format
	 * \param blob void pointer to a data buffer where the data will be stored*
	 * \param data void pointer that points to a data buffer with the data to be sent
	 * \return 0 if successful, otherwise  a TINE completion code which can be interpreted by a call to
	 * GetLastLinkError().
	 */
	int tputget(string address, string property, int size, int format, void * blob, void *data) const;

	/**
	 * \brief Client put get call
	 * This routine sends the given data to TINE and receives data from the same property
	 * \param address string with the TINE address
	 * \param property string with the TINE property
	 * \param size1 int with the size of the data input
	 * \param format1 int with the format input
	 * \param blob void pointer to a data buffer where the data will be stored
	 * \param size2 int with the size of the data output
	 * \param format2 int with the format output
	* \param data void pointer that points to a data buffer with the data to be sent
	* \return 0 if successful, otherwise  a TINE completion code which can be interpreted by a call to
	* GetLastLinkError().
	*/
	int tsendrecv(string address, string property, int size1, int format1, void * blob, int size2, int format2, void * data) const;

	/**
	 * @brief Client async get call
	 * This routine gets the property data from TINE in asynchronous mode
	 * @param address string with the TINE address
	 * @param property string with the TINE property
	 * @param size int with the size of the data
	 * @param format int with the format
	 * @param buf void pointer to a data buffer where the data will be stored
	 * @param func void pointer to the callback function (int, int)
	 * @param timeout timeout of the call
	 * @param mode (see TINE modes)
	 * @param identifier
	 * @return 0 if successful, otherwise  a TINE completion code which can be interpreted by a call to
	 * @see TINE AttachLinkEx()
	 * GetLastLinkError().
	 */
	int tgetAsync(string address, string property, int size, int format, void * buf, void (*func)(int cc,int mode), int timeout, int mode) const;
	int tgetAsyncEx(string address, string property, int size, int format, void * buf, void (*func)(int cc,int mode), int timeout, int mode, int identifier) const;

	/**
	 * \brief Client get structure call
	 * This routine gets a defined structure from TINE
	 * \param address string with the TINE address
	 * \param property string with the TINE property
	 * \param size int with the size of the data
	 * \param format int with the format
	 * \param buf void pointer to a data buffer where the data will be stored
	 * \param tag string with the name of the structure
	 * \param timeout timeout of the call
	 * \return 0 if successful, otherwise  a TINE completion code which can be interpreted by a call to
	 * GetLastLinkError().
	 */
	int tgetStruct(string address, string property, int size, int format, void * buf, int timeout, string tag) const;

	/**
	 * \brief Client put structure call
	 * This routine sends a defined structure to TINE
	 * \param address string with the TINE address
	 * \param property string with the TINE property
	 * \param size int with the size of the data
	 * \param format int with the format
	 * \param data void pointer that points to a data buffer with the data to be sent
	 * \param tag string with the name of the structure
	 * \return 0 if successful, otherwise  a TINE completion code which can be interpreted by a call to
	 * GetLastLinkError().
	 */
	int tputStruct(string address, string property, int size, int format, void * data, string tag) const;

	/**
	 * \brief Client get structure call
	 * This routine puts a defined structure and gets data from TINE
	 * \param address string with the TINE address
	 * \param property string with the TINE property
	 * \param size int with the size of the data
	 * \param format int with the format
	 * \param blob void pointer to a data buffer where the data will be stored
	 * \param data void pointer that points to a data buffer with the data to be sent
	 * \return 0 if successful, otherwise  a TINE completion code which can be interpreted by a call to
	 * GetLastLinkError().
	 */
	int tputgetStruct(string address, string property, int size, int format, void * blob, void * data) const;

	/**
	* @brief Sets the default TINE address
	* @param address pointer to char with the TINE address
	*/
    int setAddress(char * address);

	/**
	* @brief Returns the default TINE address
	* @return address pointer to char with the TINE address
	*/
    const char * getAddress();

	/**
	* @brief Sets the default TINE timeout
	* @param timeout integer with the TINE timeout
	*/
    int setTimeout(int timeout);

	/**
	* @brief Gets the default TINE timeout
	* @return timeoput integer with the TINE timeout
	*/
    int getTimeout();

	/**
	* @brief Sets the default TINE address and timeout
    * @param address pointer to char with the TINE address
	* @param timeout integer with the TINE timeout
	*/
    int initConnection(char * address, int timeout);

private:
	static PyTine *m_instance;
	static string address;
	static int timeout;
	PyTine();
	virtual ~PyTine();
    PyTine(const PyTine&);
    PyTine& operator=(const PyTine&);

};

#endif /*PYTINE_H_*/
