/** \file PyTine_py.h
 * \brief TINE PyTine API Routines
 *
 * Below are the principal PyTine API routines which allow the user to connect the Python
 * programming language with TINE control system.
 * Data acquistion can either be synchronous where execution is blocked until a call completes,
 * or asynchronous where  the results of a call are given to a callback routine upon completion.
 *
 * Note 10/08/2009: the first version implements the TINE client-side
 */


#ifndef PYTINE_PY_H_
#define PYTINE_PY_H_

#include <Python.h>
#include "PyTine.h"
#include <iostream>
#include <sstream>
#include <string>

using namespace std;

#define MAXIMUM_SIZE 8000
#define MAXIMUM_SIZE_STRUCT 2000
#define MAXIMUM_SIZE_ASYNC 1000
#define HANDLER_ARRAY_SIZE 100

//Global Variables declaration
bool initialized = false;
double *blob_async;
double blob_struct[MAXIMUM_SIZE_STRUCT];
double blob[MAXIMUM_SIZE];
double blob_aux[MAXIMUM_SIZE];
short *sval = (short *)blob;
UINT32 *lval = (UINT32 *)blob;
float *fval = (float *)blob;
double *dval = blob;
char *cval = (char *)blob;
NAME16 *n16val = (NAME16 *)blob;
NAME32 *n32val = (NAME32 *)blob;
NAME48 *n48val = (NAME48 *)blob;
NAME64 *n64val = (NAME64 *)blob;
NAME64DBLDBL *structFormat = (NAME64DBLDBL *)blob_struct;

short *struct_sval = (short *)blob_aux;
UINT32 *struct_lval = (UINT32 *)blob_aux;
float *struct_fval = (float *)blob_aux;
double *struct_dval = blob_aux;
char *struct_cval = (char *)blob_aux;


//Callback global variable
static PyObject * Handler = NULL;
static PyObject * handlerArray[HANDLER_ARRAY_SIZE];
static int handler_list_length = 0;

//Parsing Functions
BYTE *parseInputDataString(char *str,int *fmt,int *siz);
char * pythonListToString(PyObject * pList);
PyObject * parseOutput(int fmt, int size);
PyObject * extractStructData(char * tag, int * pos);
char * extractTag(char * property);

//TINE Access Functions

/* PyTine_init() */
/**
* \brief Initializes the PyTine with a fixes TINE Address and time out
*/
static PyObject* PyTine_init(PyObject*, PyObject *args);

static PyObject* PyTine_setAddress(PyObject*, PyObject *args);
static PyObject* PyTine_getAddress(PyObject*, PyObject *args);
static PyObject* PyTine_getTimeout(PyObject*, PyObject *args);
static PyObject* PyTine_setTimeout(PyObject*, PyObject *args);
static PyObject* PyTine_tgetMain (const char* address,const char* property, const char* format, int size);
static PyObject* PyTine_tget (PyObject*, PyObject *args);
static PyObject* PyTine_tgetStruct (PyObject*, PyObject *args);
static PyObject* PyTine_tgetEx (PyObject*, PyObject *args);
static PyObject* PyTine_tputMain (const char* address, const char* property, const char* format, int size, char * datastr);
static PyObject* PyTine_tput(PyObject*, PyObject *args);
static PyObject* PyTine_tputEx (PyObject* pSelf, PyObject *args);
static PyObject* PyTine_tputStruct (PyObject*, PyObject *args);
static PyObject* PyTine_tputArrayStruct(PyObject* self, PyObject *args);
static PyObject* PyTine_tputgetMain (const char* address,const char* property, const char* format, int size, char * datastr);
static PyObject* PyTine_tputget (PyObject* pSelf, PyObject *args);
static PyObject* PyTine_tputgetEx (PyObject* pSelf, PyObject *args);
static PyObject* PyTine_tsendrecvMain (const char* address,const char* property, const char* format1, int size1, const char* format2, int size2, char * datastr);
static PyObject* PyTine_tsendrecv (PyObject* pSelf, PyObject *args);
static PyObject* PyTine_tsendrecvEx (PyObject* pSelf, PyObject *args);


//Callback TINE2Python Functions
void Route_Event(int id, int cc);
void Route_EventEx(int id, int cc);
void Trigger_TINE(int id, int cc);
void Trigger_TINEEx(int id, int cc);
static PyObject* PyTine_callbackEx (PyObject* self, PyObject *args);
static PyObject* PyTine_run (PyObject*, PyObject *args);
static PyObject* PyTine_SystemCycle (PyObject*, PyObject *args);

//New Upper Level PyTine Funtions
static PyObject* PyTine_GetProperties (PyObject*, PyObject *args);
static PyObject* PyTine_GetNumProperties (PyObject*, PyObject *args);
static PyObject* PyTine_GetSrvAddress (PyObject*, PyObject *args);
static PyObject* PyTine_GetStructFormatMain (char * tag);
static PyObject* PyTine_GetStructFormat (PyObject*, PyObject *args);

//Extra TINE functions see reference documentation
static PyObject*  PyTine_CloseGlobalLink(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_CloseNetGlobal(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_FreeAccessLock(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_GetAccessLockInformation(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_GetAccessLockStatus(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_GetAlwaysRetry(PyObject* pSelf, PyObject *args); // NO ARGS
static PyObject*  PyTine_GetClnRecvQueueDepth(PyObject* pSelf, PyObject *args); // NO ARGS
static PyObject*  PyTine_GetCompletionDataSize(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_GetCompletionDataSizeFromCallbackId(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_GetCompletionDataType(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_GetCompletionDataTypeFromCallbackId(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_GetCompletionSource(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_GetCompletionSourceFromCallbackId(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_GetCompletionStatus(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_GetCompletionStatusFromCallbackId(PyObject* pSelf, PyObject *args);
//static PyObject*  PyTine_GetConnectionTable(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_GetConnectionTableCapacity(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_GetCurrentDataStatus(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_GetCurrentDataStatusFromCallbackId(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_GetCurrentDataTimeStamp(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_GetCurrentDataTimeStampFromCallbackId(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_GetCurrentLinkStatus(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_GetCurrentLinkStatusFromCallbackId(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_GetDataFromLinkId(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_GetDataFromCallbackId(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_GetDataStamp(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_GetGlobalLinkId(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_GetGlobalLinkIdFromCallbackId(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_GetGlobalsHeartbeat(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_GetGlobalsTableCapacity(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_GetGlobalsTimeStamp(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_GetLastGlobalDataSize(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_GetLastGlobalDataSizeFromCallbackId(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_GetLastLinkError(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_GetLinkName(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_GetLinkNameFromCallbackId(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_GetLinkQueueDepth(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_GetSynchronizationTolerance(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_GetSystemDataStamp(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_GetSystemDataStampFromCallbackId(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_GetTransferFlag(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_GetTransferFlagFromCallbackId(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_ModifyLinkAttributes(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_recvNetGlobal(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_recvNetGlobalEx(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_recvNetGlobalEx2(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_RegisterCycleTriggerFunction(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_SetAccessLock(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_SetAlwaysRetry(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_SetAutoLinkWatchdogs(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_SetClnRecvQueueDepth(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_SetConnectionTableCapacity(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_SetGlobalsHeartbeat(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_SetGlobalsTableCapacity(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_SetLinkQueueDepth(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_SetLinkWatchdogPollingInterval(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_SetNotificationTolerance(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_SetSuppressHeartbeatNotification(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_SetSynchronizationTolerance(PyObject* pSelf, PyObject *args);
static PyObject*  PyTine_UnregisterCycleTriggerFunction(PyObject* pSelf, PyObject *args);

// C2Python Generic Functions
PyMODINIT_FUNC initPyTine (void);

/*
 * Name: C2Python Methods
 * Description: Creates the python methods linked with the C functions
 */
static PyMethodDef PyTine_methods[] =
{
  { "pyinit", PyTine_init, METH_VARARGS, "Init the PyTine connection (string address,integer timeout)" },
  { "pygetAddress", PyTine_getAddress, METH_NOARGS, "Returns the PyTine TINE address" },
  { "pysetAddress", PyTine_setAddress, METH_VARARGS, "Sets the PyTine TINE address (string address)" },
  { "pygetTimeout", PyTine_getTimeout, METH_NOARGS, "Reads the PyTine TINE timeout" },
  { "pysetTimeout", PyTine_setTimeout, METH_VARARGS, "Sets the PyTine TINE timeout (integer timeout)" },
  { "pyget", PyTine_tget, METH_VARARGS, "Python binding for the tget function (string property,integer size,string format)" },
  { "pygetStruct", PyTine_tgetStruct, METH_VARARGS, "Python binding for getting a structure (string property)" },
  { "pygetEx", PyTine_tgetEx, METH_VARARGS, "Python binding for the tget extended function (string address,string property,integer size,string format)" },
  { "pyput", PyTine_tput, METH_VARARGS, "Python binding for the tput function (string property,string format,list data)" },
  { "pyputEx", PyTine_tputEx, METH_VARARGS, "Python binding for the tput function (string address, string property,string format,List)" },
  { "pyputget", PyTine_tputget, METH_VARARGS, "Python binding for the tgetput function (string property,string format,list data)" },
  { "pyputgetEx", PyTine_tputgetEx, METH_VARARGS, "Python binding for the tgetput extended function (string address, string property,string format,List)" },
  { "pysendrecv", PyTine_tsendrecv, METH_VARARGS, "Python binding for the tsendrecv function (string property,string formatIn,int sizeIn, string formatOut, int sizeOut, dataList)" },
  { "pysendrecvEx", PyTine_tsendrecvEx, METH_VARARGS, "Python binding for the tsendrecv extended function (string address, string property,string formatIn,int sizeIn, string formatOut, int sizeOut, dataList)" },
  { "pyputStruct", PyTine_tputStruct, METH_VARARGS, "Python binding for the tput function of a structure (string property, list data)" },
  { "pyputArrayStruct", PyTine_tputArrayStruct, METH_VARARGS, "Python binding for the tput function of an array of structures (string property,list list) (under test) " },
  { "pygetAsync", PyTine_callbackEx, METH_VARARGS, "Python binding for the async get function (string property,integer size,string format,string mode, funcHandler)" },
  { "pyrun", PyTine_run, METH_NOARGS, "Calls the TINE systemCycle" },
  { "pyCycle", PyTine_SystemCycle, METH_NOARGS, "Calls the TINE systemCycle" },
  { "pycloseGlobalLink", PyTine_CloseGlobalLink, METH_VARARGS, "Closes an active globals data link" },
  { "pycloseNetGlobal", PyTine_CloseNetGlobal, METH_VARARGS, "Closes an active globals data link according to the globals Keyword" },
  { "pygetFreeAccessLock", PyTine_FreeAccessLock, METH_VARARGS, "Frees an access lock on the server specified" },
  { "pygetAccessLockInformation", PyTine_GetAccessLockInformation, METH_VARARGS, "Acquires access lock information from the server specified" },
  { "pygetAccessLockStatus", PyTine_GetAccessLockStatus, METH_VARARGS, "Acquires access lock information from the server specified" },
  { "pygetAlwaysRetry", PyTine_GetAlwaysRetry, METH_NOARGS, "Gets the current setting of the 'Always Retry' flag" },
  { "pygetClnRecvQueueDepth", PyTine_GetClnRecvQueueDepth, METH_VARARGS, "Gets the default client-side receive queue depth for all client links" },
  { "pygetCompletionDataSize", PyTine_GetCompletionDataSize, METH_VARARGS, "Returns the most recent data size of the link index supplied" },
  { "pygetCompletionDataSizeFromCallbackId", PyTine_GetCompletionDataSizeFromCallbackId, METH_VARARGS, "Returns the most recent data size of the link with the callback id supplied" },
  { "pygetCompletionDataType", PyTine_GetCompletionDataType, METH_VARARGS, "Returns the most recent data type of the link index supplied" },
  { "pygetCompletionDataTypeFromCallbackId", PyTine_GetCompletionDataTypeFromCallbackId, METH_VARARGS, "Returns the most recent data type of the link associated with the callback id supplied" },
  { "pygetCompletionSource", PyTine_GetCompletionSource, METH_VARARGS, "Returns the most recent data type of the link associated with the callback id supplied" },
  { "pygetCompletionSourceFromCallbackId", PyTine_GetCompletionSourceFromCallbackId, METH_VARARGS, "Returns the most recent data type of the link associated with the callback id supplied" },
  { "pygetCompletionStatus", PyTine_GetCompletionStatus, METH_VARARGS, "Returns the error string associated with the input link index" },
  { "pygetCompletionStatusFromCallbackId", PyTine_GetCompletionStatusFromCallbackId, METH_VARARGS, "Returns the error string associated with the callback identifier supplied"},
  { "pygetConnectionTableCapacity", PyTine_GetConnectionTableCapacity, METH_NOARGS, "Returns the error string associated with the callback identifier supplied"},
  { "pygetCurrentDataStatus", PyTine_GetCurrentDataStatus, METH_VARARGS, "Returns the data status code associated with the input link index"},
  { "pygetCurrentDataStatusFromCallbackId", PyTine_GetCurrentDataStatusFromCallbackId, METH_VARARGS, "Returns the data status code associated with the input link index"},
  { "pygetCurrentDataTimeStamp", PyTine_GetCurrentDataTimeStamp, METH_VARARGS, "Returns the data timestamp associated with the input link indexs"},
  { "pygetCurrentDataTimeStampFromCallbackId", PyTine_GetCurrentDataTimeStampFromCallbackId, METH_VARARGS, "Returns the data timestamp associated with the input link indexs"},
  { "pygetCurrentLinkStatus", PyTine_GetCurrentLinkStatus, METH_VARARGS, "Returns the completion code associated with the input link index"},
  { "pygetCurrentLinkStatusFromCallbackId", PyTine_GetCurrentLinkStatusFromCallbackId, METH_VARARGS, "Returns the completion code associated with the callback identifier supplied"},
  { "pygetDataFromLinkId", PyTine_GetDataFromLinkId, METH_VARARGS, "Supplies the DTYPE data object and call status for the link ID in question"},
  { "pygetDataFromCallbackId", PyTine_GetDataFromCallbackId, METH_VARARGS, "Supplies the DTYPE data object and call status for the Callback ID in question"},
  { "pygetDataStamp", PyTine_GetDataStamp, METH_VARARGS, "Gets the data stamp with which the incoming data set has been tagged"},
  { "pygetGlobalLinkId", PyTine_GetGlobalLinkId, METH_VARARGS, "Retrieves the link ID for a globals link"},
  { "pygetGlobalLinkIdFromCallbaclId", PyTine_GetGlobalLinkIdFromCallbackId, METH_VARARGS, "retrieves the link ID for a globals link"},
  { "pygetGlobalsHeartbeat", PyTine_GetGlobalsHeartbeat, METH_NOARGS, "gets the heartbeat (in seconds) during which time no 'timeout' messages will be sent to a globals dispatch routine (client-side)"},
  { "pygetGlobalsTableCapacity", PyTine_GetGlobalsTableCapacity, METH_NOARGS, "gets the globals table capacity (client-side)"},
  { "pygetGlobalsTimeStamp", PyTine_GetGlobalsTimeStamp, METH_VARARGS, "Returns the timestamp of the globals keyword with the given link ID"},
  { "pygetLastGlobalDataSize", PyTine_GetLastGlobalDataSize, METH_VARARGS, "Returns the most recent data size of the global link index supplied"},
  { "pygetLastGlobalDataSizeFromCallbackId", PyTine_GetLastGlobalDataSizeFromCallbackId, METH_VARARGS, "Returns the most recent data size of the global callback id supplied"},
  { "pygetLastLinkError", PyTine_GetLastLinkError, METH_VARARGS, "The error string associated with the input error number"},
  { "pygetLinkName", PyTine_GetLinkName, METH_VARARGS, "Returns the full link name associated with the input link index"},
  { "pygetLinkFromCallbackId", PyTine_GetLinkNameFromCallbackId, METH_VARARGS, "Returns the full link name associated with the callback identifier supplied"},
  { "pygetLinkQueueDepth", PyTine_GetLinkQueueDepth, METH_VARARGS, "Gets the client-side receive queue depth for the link in question"},
  { "pygetSynchronizationTolerance", PyTine_GetSynchronizationTolerance, METH_NOARGS, "Gets the tolerance used in deciding whether to apply a timestamp offset or not"},
  { "pygetSystemDataStamp", PyTine_GetSystemDataStamp, METH_VARARGS, "Gets the global system data stamp with which the incoming data set has been tagged"},
  { "pygetSystemDataStampFromCallbackId", PyTine_GetSystemDataStampFromCallbackId, METH_VARARGS, "Gets the global system data stamp with which the incoming data set has been tagged"},
  { "pygetTransferFlag", PyTine_GetTransferFlag, METH_VARARGS, "Gets the data transfer flag for the given link ID"},
  { "pygetTransferFlagFromCallbackId", PyTine_GetTransferFlagFromCallbackId, METH_VARARGS, "Gets the data transfer flag for the given callback id"},
  { "pyModifyLinkAttributes", PyTine_ModifyLinkAttributes, METH_VARARGS, "Allows the caller to assign new link attributes to an active link"},
  { "pyrecvNetGlobal", PyTine_recvNetGlobal, METH_VARARGS, "Initiates a net globals data link"},
  { "pyrecvNetGlobalEx", PyTine_recvNetGlobalEx, METH_VARARGS, "Initiates a net globals data link (extended call)"},
  { "pyrecvNetGlobalEx2", PyTine_recvNetGlobalEx2, METH_VARARGS, "Initiates a net globals data link (extended call)"},
  { "pyregisterCycleTriggerFunction", PyTine_RegisterCycleTriggerFunction, METH_VARARGS, "Registers a cycle trigger callback dispatch function (not implemented)"},
  { "pysetAccessLock", PyTine_SetAccessLock, METH_VARARGS, "Acquires an access lock to the server specified"},
  { "pysetAlwaysRetry", PyTine_SetAlwaysRetry, METH_VARARGS, "Sets the 'Always Retry' flag to the value given"},
  { "pysetAutoLinkWatchdogs", PyTine_SetAutoLinkWatchdogs, METH_VARARGS, "Enables/Disables automatic link watchdogs"},
  { "pysetClnRecvQueueDepth", PyTine_SetClnRecvQueueDepth, METH_VARARGS, "Sets the default client-side receive queue depth for all client links"},
  { "pysetConnectionTableCapacity", PyTine_SetConnectionTableCapacity, METH_VARARGS, "Sets the maximum number of entries in the connection table"},
  { "pysetGlobalsHeartbeat", PyTine_SetGlobalsHeartbeat, METH_VARARGS, "sets the heartbeat (in seconds) during which time no 'timeout' messages will be sent to a globals dispatch routine (client-side)"},
  { "pysetGlobalsTableCapacity", PyTine_SetGlobalsTableCapacity, METH_VARARGS, "Sets the globals table capacity"},
  { "pysetLinkQueueDepth", PyTine_SetLinkQueueDepth, METH_VARARGS, "Sets the client-side receive queue depth for the link in question"},
  { "pysetLinkWatchdogPollingInterval", PyTine_SetLinkWatchdogPollingInterval, METH_VARARGS, "Sets the link watchdog polling interval to the value given"},
  { "pysetNotificationTolerance", PyTine_SetNotificationTolerance, METH_VARARGS, "Allows the caller to apply a tolerance pertaining to link notification"},
  { "pysetSuppressHeartbeatNotification", PyTine_SetSuppressHeartbeatNotification, METH_VARARGS, "Determines whether CM_DATACHANGE data links signal incoming data by calling the corresonding callback routine even when the incoming data is a HEARTBEAT notification"},
  { "pysetSynchronizationTolerance", PyTine_SetSynchronizationTolerance, METH_VARARGS, "Sets the tolerance used in deciding whether to apply a timestamp offset or not"},
  { "pyunregisterCycleTriggerFunction", PyTine_UnregisterCycleTriggerFunction, METH_VARARGS, "Unregisters a previously registered cycle trigger callback dispatch function"},
  { "pygetProperties", PyTine_GetProperties, METH_NOARGS, "Returns the properties of a server"},
  { "pygetNumProperties", PyTine_GetNumProperties, METH_NOARGS, "Returns the number of properties of a server"},
  { "pygetSrvAddress", PyTine_GetSrvAddress, METH_NOARGS, "Returns the server address information"},
  { "pygetStructFormat", PyTine_GetStructFormat, METH_VARARGS, "Returns the format of a given structure (string StructTag)"},
  {NULL, NULL, 0, NULL}

};



#endif /*PYTINE_PY_H_*/
